/* Fiasco - computes sample statistics.
   Copyright (C) 1997 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   The author may be contacted at <pfaffben@pilot.msu.edu> on the
   Internet, or as Ben Pfaff, 12167 Airport Rd, DeWitt MI 48820, USA
   through more mundane means. */

#if !tokens_h
#define tokens_h 1

#include "str.h"
#include "common.h"

/* Returns nonzero if character CH may be the first character in an
   identifier. */
#define is_id1(CH) 							\
	(isalpha (CH) || (CH) == '@' || (CH) == '#' || (CH) == '$')

/* Returns nonzero if character CH may be a character in an
   identifier other than the first. */
#define is_idn(CH)					\
	(is_id1 (CH) || (CH) == '.' || (CH) == '_')

/* Token types. */
enum
  {
    /* Basic tokens. */
    ID = 256, NUM, STRING, STOP,

    /* Keywords.  The order is important here. */
    AND, OR, NOT,
    EQ, GE, GT, LE, LT, NE,	/* Do not separate the relational operators. */
    ALL, BY, TO, WITH, EXP,	/* Do not separate the ** operator. */

    /* Tagged quotes. */
    TAGGED_QUOTE,
  };

extern int token;		/* Current token. */
extern double tokval;		/* NUM token: token's value. */
extern unsigned char *tokstr;	/* ID, STRING, TAGGED_QUOTE: token's value. */
extern int tokstr_len;		/* TAGGED_QUOTE: tokstr's length. */
extern int tokstr_size;		/* Number of bytes allocated for tokstr. */
extern unsigned char *toklongstr;	/* ID token: full id, not truncated to 8 */
extern long int tokint;		/* NUM token: token's value if integer,
				   NOT_LONG otherwise */

/* Main lexer routines. */
int dump_token (void);
int yylex (void);
void include (char *fname);

/* Functions for coping with odd syntactic situations. */
int lookahead (void);
void putback (int);
char *get_rest_of_line (void);
char *get_dotted_rest_of_line (int *had_dot);
char *get_entire_line (void);
void discard_line (void);
char *tokname (int);
int preprocess_line (void);
char *get_token_representation (void);
void convert_negative_to_dash (void);

#if PRODUCTION
#define get_token() (token = yylex())
#else
#define get_token() (dump_token())
#endif

int match_id (char *);
int match_tok (int);
int match_int (int);

#if __GNUC__
/* If the current token is the identifier S (specified without
   quotes), skips the token and returns 1; otherwise, returns 0. */
#define match_id(S)						\
	({							\
	  int x = (token == ID && id_match (#S, tokstr));	\
	  if (x)						\
	    get_token ();					\
	  x;							\
	})

/* If TOK is the current token, skips it and returns 1; otherwise,
   returns 0. */
#define match_tok(TOK)				\
	({					\
	  int x = (token == TOK);		\
	  if (x)				\
	    get_token ();			\
	  x;					\
	})
	
/* If the current token is an integer having value N, skips it and
   returns 1; otherwise, returns 0. */
#define match_int(N)					\
	({						\
	  int x = (token == NUM && tokint == N);	\
	  if (x)					\
	    get_token ();				\
	  x;						\
	})
#else /* !__GNUC__ */
#define match_id(S)				\
	match_id (#S)
#endif /* !__GNUC__ */

/* If the current token is the identifier S (specified without
   quotes), skips the token; otherwise, reports an error and returns
   from the current function with return value 0. */
#define force_match_id(S)					\
	if (token != ID || !id_match (#S, tokstr))		\
	  return msg (SE, "`" #S " %s", _("' expected"));	\
	else							\
	  get_token ()
	
/* If the current token is T, skips the token; otherwise, reports an
   error and returns from the current function with return value 0. */
#define force_match(T)					\
	if (token != T)					\
	  return msg (SE, #T "%s", _(" expected"));	\
	else						\
	  get_token ()
	
/* If the current token is not a string, reports an error and returns
   from the current function with return value 0. */
#define force_string() 					\
	do						\
	  {						\
	    if (token != STRING)			\
	      return msg (SE, _("String expected."));	\
	  }						\
	while (0)

/* If the current token is not an integer, reports an error and
   returns from the current function with return value 0. */
#define force_int()					\
	do						\
	  {						\
	    if (token != NUM || tokint == NOT_LONG)	\
	      return msg (SE, _("Integer expected."));	\
	  }						\
	while (0)
	
/* If the current token is not a number, reports an error and returns
   from the current function with return value 0. */
#define force_num() 					\
	do						\
	  {						\
	    if (token != NUM)				\
	      return msg (SE, _("Number expected."));	\
	  }						\
	while (0)
	
/* If the current token is not an identifier, reports an error and
   returns from the current function with return value 0. */
#define force_id() 						\
	do							\
	  {							\
	    if (token != ID)					\
	      return msg (SE, _("Variable name expected."));	\
	  }							\
	while (0)

int id_match (char *keyword, char *token);

#endif /* tokens_h */
